<html>
<head>
  <title>My Bookmarks</title>
</head>
<body>
  <h1>My Bookmarks</h1>

<dl>
{% for entry in entries %}
  <dd><a href="{{ entry.data.url }}">{{ entry.data.title }}</a></dd>
  <dl>
    <div>{{ entry.data.description }}<div>
    <div>{{ entry.data.added }}</div>
    <ul>
    {% for t in entry.data.tags %}
      <li>{{ t }}
    {% endfor %}
    </ul>
  </dl>
{% endfor %}
</dl>
</body>
</html>
